import os

DEFAULT_NAMESPACE = os.getenv('KUBE_NAMESPACE', 'default')
DEFAULT_SERVICE_ACCOUNT_NAME = os.getenv('KUBE_SERVICE_ACCOUNT_NAME', 'default')
DEFAULT_STORAGE_CLASS_NAME = os.getenv('KUBE_STORAGE_CLASS_NAME', 'default')

KUBE_POD_NAME_ENV_VAR = 'HOSTNAME'
