CUSTOM_TEMPLATES_DIRECTORY = 'custom_templates'
CUSTOM_TEMPLATES_DIRECTORY_ENVIRONMENT_VARIABLE = 'CUSTOM_TEMPLATES_DIRECTORY'

DIRECTORY_FOR_BLOCK_TEMPLATES = 'blocks'
DIRECTORY_FOR_PIPELINE_TEMPLATES = 'pipelines'

METADATA_FILENAME_WITH_EXTENSION = 'metadata.yaml'
