(self.webpackChunk_N_E=self.webpackChunk_N_E||[]).push([[3277],{84392:function(e,n,r){"use strict";r.d(n,{HF:function(){return i},L6:function(){return s}});var s,t=r(46336),u=r(22341);function i(e,n,r){var i=e.owner,a=(e.roles,[{Icon:u.Vz,id:s.WORKSPACES,isSelected:function(){return s.WORKSPACES===r},label:function(){return"Workspaces"},linkProps:{href:"/manage"}}]);return i&&a.push({Icon:u.NO,id:s.USERS,isSelected:function(){return s.USERS===r},label:function(){return"Users"},linkProps:{href:"/manage/users"}}),n==t.k.MAIN&&a.push({Icon:u.Zr,id:s.SETTINGS,isSelected:function(){return s.SETTINGS===r},label:function(){return"Settings"},linkProps:{href:"/manage/settings"}}),a}!function(e){e.WORKSPACES="workspaces",e.USERS="users",e.SETTINGS="settings"}(s||(s={}))},3849:function(e,n,r){"use strict";var s=r(82684),t=r(1210),u=r(82531),i=r(49125),a=r(84392),o=r(9736),c=r(28598);n.Z=function(e){var n=e.before,r=e.breadcrumbs,l=void 0===r?[]:r,d=e.children,f=e.errors,p=e.pageName,S=e.subheaderChildren,h=u.ZP.statuses.list().data,b=(0,s.useMemo)((function(){var e,n;return null===h||void 0===h||null===(e=h.statuses)||void 0===e||null===(n=e[0])||void 0===n?void 0:n.project_type}),[h]),v=(0,o.PR)()||{};return(0,c.jsx)(t.Z,{before:n,beforeWidth:n?50*i.iI:0,breadcrumbs:l,errors:f,navigationItems:(0,a.HF)(v,b,p),subheaderChildren:S,title:"Workspaces",uuid:"workspaces/index",children:d})}},29273:function(e,n,r){"use strict";r.r(n);var s=r(77837),t=r(38860),u=r.n(t),i=(r(82684),r(34376)),a=r(41788),o=r(86673),c=r(20582),l=r(3849),d=r(49125),f=r(84392),p=r(28598);function S(){var e=(0,i.useRouter)();return(0,p.jsx)(l.Z,{breadcrumbs:[{label:function(){return"Workspaces"},linkProps:{as:"/manage",href:"/manage"}},{label:function(){return"Users"},linkProps:{as:"/manage/users",href:"/manage/users"}},{bold:!0,label:function(){return"New"}}],pageName:f.L6.USERS,children:(0,p.jsx)(o.Z,{p:d.cd,children:(0,p.jsx)(c.Z,{hideFields:["roles"],newUser:!0,onSaveSuccess:function(n){e.push("/manage/users/[user]","/manage/users/".concat(null===n||void 0===n?void 0:n.id))},title:"Add new user",user:{}})})})}S.getInitialProps=function(){var e=(0,s.Z)(u().mark((function e(n){return u().wrap((function(e){for(;;)switch(e.prev=e.next){case 0:return e.abrupt("return",{});case 1:case"end":return e.stop()}}),e)})));return function(n){return e.apply(this,arguments)}}(),n.default=(0,a.Z)(S)},47272:function(e,n,r){(window.__NEXT_P=window.__NEXT_P||[]).push(["/manage/users/new",function(){return r(29273)}])}},function(e){e.O(0,[844,1424,1005,3881,9774,2888,179],(function(){return n=47272,e(e.s=n);var n}));var n=e.O();_N_E=n}]);