ACTIVE_DIRECTORY_CLIENT_ID = '51aec820-9d49-40a9-b046-17c1f28f620d'

GITHUB_CLIENT_ID = '8577f13ddc81e2848b07'
GITHUB_STATE = '1337'

OAUTH_PROVIDER_ACTIVE_DIRECTORY = 'active_directory'
OAUTH_PROVIDER_GITHUB = 'github'

VALID_OAUTH_PROVIDERS = [
    OAUTH_PROVIDER_ACTIVE_DIRECTORY,
    OAUTH_PROVIDER_GITHUB,
]
